import copy
import functools
import typing as t


class Dependency:
    @classmethod
    def add(cls, name: str, service: t.Any) -> None:
        setattr(cls, f"__{name}", service)

    @classmethod
    def override(cls, name: str, service: t.Any) -> None:
        setattr(cls, f"__{name}", service)

    @classmethod
    def inject(cls, name: str) -> t.Any:
        def decorator(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                args = args + (copy.copy(getattr(cls, f"__{name}")),)
                return await func(*args, **kwargs)

            return wrapper

        return decorator



import copy
import typing as t

from pydantic import BaseModel

from . import construct


class Spec:
    def __init__(self) -> None:
        self.methods: t.List[construct.Method] = []
        self.definition = construct.Definition(paths={}, components=None)

    def add_openapi(self, version: str) -> None:
        self.definition.openapi = version

    def add_info(self, title: str, version: str, description: str = None) -> None:
        self.definition.info = construct.Info(title=title, version=version, description=description)

    def update_tag(self, name: str, description: str) -> None:
        if self.definition.tags is None:
            return None

        for tag in self.definition.tags:
            if name == tag.name:
                tag.description = description

    def add_endpoint(
        self,
        classname: str,
        methodname: str,
        tags: t.Optional[t.List[str]] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        parameters: t.List[construct.Parameter] = None,
        body: construct.RequestBody = None,
        response: t.Dict[str, construct.Response] = None,
    ) -> None:
        method = self._get_method(classname, methodname)
        self._add_tags(tags)
        if isinstance(method, construct.Method):
            endpoint = self._update_endpoint(method.endpoint, tags, summary, description, parameters, body, response)
            self._update_method(classname, methodname, endpoint)
            return None

        self._add_method(classname, methodname, tags, summary, description, parameters, body, response)

    def to_parameters(self, serializer: BaseModel, in_: str) -> t.List[construct.Parameter]:
        parameters: t.List[construct.Parameter] = []
        serializer = serializer.schema()
        required = serializer.get("required") or []
        for key, value in serializer["properties"].items():
            parameter = construct.Parameter(
                name=key, in_=in_, required=True if key in required else False, schema_=value
            )
            parameters.append(parameter)

        return parameters

    def to_body(self, serializer: BaseModel, description: str, content_type: str) -> construct.RequestBody:
        content = construct.Content(schema_={"$ref": f"#/components/schemas/{serializer.__name__}"})
        self._update_schema(serializer)
        return construct.RequestBody(description=description, content={content_type: content})

    def to_response(
        self, serializer: BaseModel, description: str, accept: str = "application/json"
    ) -> construct.Response:
        content = construct.Content(schema_={"$ref": f"#/components/schemas/{serializer.__name__}"})
        self._update_schema(serializer)
        return construct.Response(description=description, content={accept: content})

    def update_definition(self, classname: str, pathname: str, routing_methods: set = None) -> None:
        for method in self.methods:
            if routing_methods:
                if method.name.upper() in routing_methods and classname == method.classname:
                    self._update_paths(method, pathname)

                continue
            else:
                if classname == method.classname:
                    self._update_paths(method, pathname)

    def _get_method(self, classname: str, methodname: str) -> t.Union[construct.Method, None]:
        for method in self.methods:
            if classname == method.classname and methodname == method.name:
                return method

        return None

    def _add_method(
        self,
        classname: str,
        methodname: str,
        tags: t.Optional[t.List[str]],
        summary: t.Optional[str],
        description: t.Optional[str],
        parameters: t.List[construct.Parameter] = None,
        body: construct.RequestBody = None,
        response: t.Dict[str, construct.Response] = None,
    ) -> None:
        endpoint = construct.Endpoint(summary=summary, description=description)
        if tags:
            self._add_tags(tags)
            endpoint.tags = tags

        if parameters:
            endpoint.parameters = parameters

        if body:
            endpoint.requestBody = body

        if response:
            endpoint.responses = response

        method = construct.Method(name=methodname, classname=classname, endpoint=endpoint)
        self.methods.append(method)

    def _update_method(self, classname: str, methodname: str, endpoint: construct.Endpoint) -> None:
        for i, method in enumerate(self.methods):
            if classname == method.classname and methodname == method.name:
                method.endpoint = endpoint
                self.methods[i] = method

    def _update_schema(self, serializer: BaseModel) -> None:
        if serializer.schema().get("definitions"):
            self._update_nested_schema(serializer)
            return None

        self._update_schemas({serializer.__name__: serializer.schema()})

    def _update_nested_schema(self, serializer: BaseModel) -> None:
        schema = copy.copy(serializer.schema(ref_template="#/components/schemas/{model}"))
        definitions = schema.pop("definitions")
        self._update_schemas({serializer.__name__: schema})
        for key, value in definitions.items():
            self._update_schemas({key: value})

    def _update_schemas(self, schema: dict) -> None:
        if self.definition.components is None:
            self.definition.components = construct.Schemas(schemas=schema)
            return None

        self.definition.components.schemas.update(schema)

    def _update_endpoint(
        self,
        endpoint: construct.Endpoint,
        tags: t.Optional[t.List[str]],
        summary: t.Optional[str],
        description: t.Optional[str],
        parameters: t.List[construct.Parameter] = None,
        body: t.Dict[str, construct.RequestBody] = None,
        response: t.Dict[str, construct.Response] = None,
    ) -> construct.Endpoint:
        if tags:
            endpoint.tags = tags

        if summary:
            endpoint.summary = summary

        if description:
            endpoint.description = description

        if parameters:
            if endpoint.parameters is None:
                endpoint.parameters = []

            endpoint.parameters += parameters

        if body:
            endpoint.requestBody = body

        if response:
            if endpoint.responses is None:
                endpoint.responses = {}

            endpoint.responses.update(response)

        return endpoint

    def _update_paths(self, method: construct.Method, pathname: str) -> None:
        path = construct.Path(name=pathname, method={method.name: method.endpoint})
        current_path_method = self.definition.paths.get(path.name)
        if current_path_method:
            current_path_method.update(path.method)
            self.definition.paths.update({path.name: current_path_method})
            return None

        self.definition.paths.update({path.name: path.method})

    def _add_tags(self, tags: t.List[str]) -> None:
        if tags is None:
            return None

        if self.definition.tags is None:
            self.definition.tags = [construct.Tag(name=tag) for tag in tags]
            return None

        current_tags: t.List[str] = [current_tag.name for current_tag in self.definition.tags]
        for tag in tags:
            if tag not in current_tags:
                self.definition.tags.append(construct.Tag(name=tag))



import typing as t

import jinja2
from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse, JSONResponse
from starlette.routing import BaseRoute, Mount, Route

from .api import Api


class OpenApiEndpoint(HTTPEndpoint):
    async def get(self, request):
        def updates(routes: t.List[BaseRoute], parent: str):
            for route in routes:
                if isinstance(route, Route):
                    Api.spec.update_definition(route.name, f"{parent}{route.path}", route.methods)

                if isinstance(route, Mount):
                    updates(route.routes, f"{parent}{route.path}")

        updates(request.app.routes, "")

        return JSONResponse(Api.spec.definition.dict(by_alias=True, exclude_none=True))


class SwaggerUiEndpoint(HTTPEndpoint):
    async def get(self, request):
        environment = jinja2.Environment()
        template = environment.from_string(
            """<!DOCTYPE html>
            <html lang="en">

            <head>
                <meta charset="utf-8" />
                <meta name="viewport" content="width=device-width, initial-scale=1" />
                <meta name="description" content="SwaggerUI" />
                <title>SwaggerUI</title>
                <link rel="stylesheet" href="https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui.css" />
            </head>

            <body>
                <div id="swagger-ui"></div>
                <script src="https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui-bundle.js" crossorigin></script>
                <script>
                    window.onload = () => {
                        window.ui = SwaggerUIBundle({
                            url: "{{ url }}",
                            dom_id: "#swagger-ui",
                        });
                    };
                </script>
            </body>

            </html>
        """
        )
        html = template.render(url=request.url_for("OpenApiEndpoint"))
        return HTMLResponse(html)


class RedocEndpoint(HTTPEndpoint):
    async def get(self, request):
        environment = jinja2.Environment()
        template = environment.from_string(
            """<!DOCTYPE html>
            <html>

            <head>
                <title>Redoc</title>
                <!-- needed for adaptive design -->
                <meta charset="utf-8"/>
                <meta name="viewport" content="width=device-width, initial-scale=1">
                <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

                <!--
                Redoc doesn't change outer page styles
                -->
                <style>
                body {
                    margin: 0;
                    padding: 0;
                }
                </style>
            </head>

            <body>
                <redoc spec-url="{{ url }}"></redoc>
                <script src="https://cdn.redoc.ly/redoc/latest/bundles/redoc.standalone.js"> </script>
            </body>
            
            </html>
        """
        )
        html = template.render(url=request.url_for("OpenApiEndpoint"))
        return HTMLResponse(html)



from starlette import datastructures


class RequestFile(bytes):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, datastructures.UploadFile):
            raise TypeError("Starlette data structures UploadFile required")
        return v

    def __repr__(self):
        return f"FilesType({super().__repr__()})"



import typing as t

from pydantic import BaseModel, Field


class Schemas(BaseModel):
    schemas: t.Dict[str, t.Any]


class Parameter(BaseModel):
    name: str
    in_: str = Field(alias="in")
    required: bool
    schema_: dict = Field(alias="schema")
    description: t.Optional[str]

    class Config:
        allow_population_by_field_name = True


class Content(BaseModel):
    schema_: t.Dict[str, str] = Field(alias="schema")

    class Config:
        allow_population_by_field_name = True


class Response(BaseModel):
    description: t.Optional[str]
    content: t.Optional[t.Dict[str, Content]]


class RequestBody(BaseModel):
    description: t.Optional[str]
    content: t.Optional[t.Dict[str, Content]]


class Endpoint(BaseModel):
    tags: t.Optional[t.List[str]]
    summary: t.Optional[str]
    description: t.Optional[str]
    requestBody: t.Optional[RequestBody]
    parameters: t.Optional[t.List[Parameter]]
    responses: t.Optional[t.Dict[str, Response]]


class Method(BaseModel):
    name: str
    classname: str
    endpoint: Endpoint


class Path(BaseModel):
    name: str
    method: t.Dict[str, Endpoint]


class Tag(BaseModel):
    name: t.Optional[str]
    description: t.Optional[str]


class Info(BaseModel):
    title: str = "StarAPI"
    version: str = "0.0.1"
    description: t.Optional[str]


class Definition(BaseModel):
    openapi: t.Optional[str]
    info: t.Optional[Info]
    tags: t.Optional[t.List[Tag]]
    paths: t.Optional[t.Dict[str, t.Any]]
    components: t.Optional[Schemas]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starapi', 'starapi.spec']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0', 'pydantic>=1.10.2,<2.0.0', 'starlette>=0.20.4,<0.21.0']

setup_kwargs = {
    'name': 'starapi',
    'version': '0.1.5',
    'description': 'StarAPI based on starlette and pydantic.',
    'long_description': '# starapi',
    'author': 'jonars',
    'author_email': 'jonarsli13@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/JonarsLi/flask-restapi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
